from .exceptions import InvalidInput


class Page:

    def __init__(self, start=None, length=None, page=None):
        if page:
            try:
                start = page["start"]
            except KeyError:
                pass
            try:
                length = page["length"]
            except KeyError:
                pass
        if start is not None and start < 1:
            raise InvalidInput("Start index cannot be less than 1")
        if length and length > 50:
            length = 50
        self._start = start or 1
        self._length = length or 10

    def get_start(self):
        return self._start

    def get_length(self):
        return self._length

    def get_end(self):
        return self._start + self.get_length() - 1

    def to_dict(self):
        return {
            "start": self.get_start(),
            "length": self.get_length()
        }


class Pagination:

    def __init__(self, current_page=None):
        if current_page:
            self._current_page = current_page
        else:
            self._current_page = Page()
        self._updated_current_page = None
        self._total_count = 0

    def get_current_page(self):
        return self._current_page

    def set_current_page(self, page):
        self._current_page = page

    def update_page_properties(self, current_page_count, total_count):
        self._updated_current_page = Page(self.get_current_page().get_start(), current_page_count)
        self._total_count = total_count

    def get_next_page(self):
        curr_end = self.get_current_page().get_end()
        curr_len = self.get_current_page().get_length()
        if curr_end < self._total_count:
            next_total = self._total_count - curr_end
            next_len = next_total if next_total < curr_len else curr_len
            return Page(curr_end+1, next_len)
        return None

    def get_previous_page(self):
        curr_start = self.get_current_page().get_start()
        if curr_start > 1:
            curr_len = self.get_current_page().get_length()
            # prev_len = curr_start if curr_start < curr_len else curr_len
            prev_len = curr_len
            prev_start = curr_start - prev_len
            if prev_start < 1:
                prev_start = 1
            return Page(prev_start, prev_len)
        return None

    def to_dict(self):
        next_page = self.get_next_page()
        if next_page:
            next_page = next_page.to_dict()
        previous_page = self.get_previous_page()
        if previous_page:
            previous_page = previous_page.to_dict()
        return {
            "previous": previous_page,
            "current": self.get_current_page().to_dict(),
            "next": next_page
        }


class Meta:

    def __init__(self, current_page=None, order_by=None, required_fields=None):
        self._pagination = Pagination(current_page)
        self._order_by = order_by
        self._required_fields = required_fields
        self._total_count = None

    def get_pagination(self):
        return self._pagination

    def get_order_by(self):
        return self._order_by

    def set_order_by(self, *order_by):
        self._order_by = order_by

    def set_required_fields(self, *fields):
        self._required_fields = fields

    def get_required_fields(self):
        return self._required_fields or ()

    def update_page_properties(self, current_page_count, total_count):
        self._total_count = total_count
        self._pagination.update_page_properties(current_page_count, total_count)

    def to_dict(self):
        return {
            "page": self.get_pagination().to_dict(),
            "total_count": self._total_count
        }



from itertools import chain
from .request import APIMethods
from django.db.models import Model, ForeignKey
from django.db.models.fields import DateTimeField
from enum import Enum
from .response_utils import ApiResponse
from .exceptions import AdharaException
from .utilities import is_reference_field


class ResponseModes(Enum):

    JSON = "json"
    MSG_PACK = "msg_pack"


class ResponseStatus(Enum):
    SUCCESS = "success"
    ERROR = "error"
    FAILURE = "failure"


class AdharaResponse:

    def __init__(self, adhara_request):
        self._request = adhara_request
        self._response_data = None
        self._status = None
        self._transformed_data = None
        self._permissions = {}
        self._error_code = 200

    def success(self, data):
        self._response_data = data
        self._status = ResponseStatus.SUCCESS

    def error(self, data, error_code=400):
        self._response_data = data
        self._status = ResponseStatus.ERROR
        self._error_code = error_code

    def failure(self, data):
        self._response_data = data
        self._status = ResponseStatus.FAILURE

    def set_permission(self, pk, name, is_permitted):
        try:
            self._permissions[pk]
        except KeyError:
            self._permissions[pk] = {}
        self._permissions[pk][name] = is_permitted is True

    def get_response_status(self):
        return self._status

    def _get_response_meta(self):
        return self._request.get_meta().to_dict()

    def _model_to_dict(self, instance, fields=None, exclude=None):
        opts = instance._meta
        data = {}
        fields = fields or self._request.get_meta().get_required_fields()
        for f in chain(opts.concrete_fields, opts.private_fields, opts.many_to_many):
            skip_serialization = False
            formatted_val = None
            if type(f) == DateTimeField:
                iso_ts = f.value_from_object(instance)
                formatted_val = iso_ts.strftime('%H:%M %p, %B %d, %Y')
            elif is_reference_field(f):
                try:
                    skip_serialization = f.model._meta.get_field(f.name).rel.to.Adhara.skip_serialization
                except AttributeError:
                    pass
                if not skip_serialization:
                    value_object = getattr(instance, f.name)
                    if value_object:
                        formatted_val = {
                            "id": value_object.pk,
                            "display_value": str(value_object)
                        }
                    else:
                        formatted_val = None
            if exclude and f.name in exclude or skip_serialization:
                continue
            if fields:
                if f.name in fields:
                    data[f.name] = formatted_val if formatted_val else f.value_from_object(instance)
            else:
                data[f.name] = formatted_val if formatted_val else f.value_from_object(instance)
        sub_requests = self._request.get_sub_requests()
        # TODO get sub request model
        if sub_requests:
            from .restview import RestView
            for sub_request in sub_requests:
                sub_request.model_manager = getattr(instance, sub_request.get_model()._meta.object_name.lower()+"_set")
                transformed_data = RestView(sub_request).process().get_transformed_data()
                if sub_request.get_response_key():
                    data[sub_request.get_response_key()] = transformed_data
                elif type(transformed_data) == dict:
                    data.update(transformed_data)
                else:
                    raise AdharaException("Response key not provided")
        self._call_transformer(instance, data)
        return data

    def transform_row(self, data, fields=None, exclude=None):
        if isinstance(data, Model):
            dict_response = self._model_to_dict(data, fields=fields, exclude=exclude)
            if self._permissions:
                try:
                    if self._permissions[data.pk]:
                        dict_response["_permissions"] = self._permissions[data.pk]
                except KeyError:
                    pass
            return dict_response
        else:
            return data

    def get_request(self):
        return self._request

    def get_data(self):
        return self._response_data

    def _call_transformer(self, row, transformed_row):
        if self._request.transformer:
            self._request.transformer(row, transformed_row)

    def transform_data(self):
        if self.get_response_status() == ResponseStatus.SUCCESS:
            # TODO integrate with model to dict
            if type(self.get_data()) == list:
                transformed_data = []
                for row in self.get_data():
                    transformed_data.append(self.transform_row(row))
            else:
                transformed_data = self.transform_row(self.get_data())
            self._transformed_data = transformed_data
        else:
            self._transformed_data = self.get_data()

    def get_transformed_data(self):
        if not self._transformed_data:
            self.transform_data()
        return self._transformed_data

    def get_response(self):
        response = {
            "data": self.get_transformed_data(),
            "status": self._status.value
        }
        if self._request.get_method() == APIMethods.GET_LIST:
            meta = self._get_response_meta()
            if meta:
                response["meta"] = meta
        return response

    def serialize(self, dict_response=None, response_mode=ResponseModes.JSON):
        dict_response = dict_response or self.get_response()
        if response_mode == ResponseModes.JSON:
            if self._status == ResponseStatus.FAILURE:
                return ApiResponse.AsiddhauPratikriyaa(dict_response)
            elif self._status == ResponseStatus.ERROR:
                if self._error_code == 403:
                    return ApiResponse.NotAllowedPratikriyaa(dict_response)
                return ApiResponse.SkhalitaPratikriyaa(dict_response)
            else:
                return ApiResponse.Pratikriyaa(dict_response)



from django.db import models
from adhara.session import Session
from ..utilities import get_config
from ..exceptions import InvalidDatabaseSelected
from ..utilities import is_reference_field


def fill_ref_models(self, request=None, **kwargs):
    if request is None:
        return kwargs
    for field in self._meta.fields:
        vn = field.name
        if vn in kwargs:
            if kwargs[vn] == "0":
                del kwargs[vn]
            elif is_reference_field(field):
                db = Session.get_db(request)
                if type(kwargs[vn]) == str or type(kwargs[vn]) == int:
                    objects = field.related_model.objects
                    if db is None:
                        kwargs[vn] = objects.get(pk=kwargs[vn])
                    else:
                        kwargs[vn] = objects.using(db).get(pk=kwargs[vn])
                elif type(kwargs[vn]) == dict:
                    args = kwargs[vn]
                    args['request'] = request
                    kwargs[vn] = field.related_model(**args)
                    if db is None:
                        kwargs[vn].save()
                    else:
                        kwargs[vn].save(using=db)
    return kwargs


class AdharaManager(models.Manager):

    def __init__(self, *args, **kwargs):
        super(AdharaManager, self).__init__(*args, **kwargs)


class AdharaModel(models.Model):

    objects = AdharaManager()

    def fill_defaults(self, request, **kwargs):
        return kwargs

    def __init__(self, *args, **kwargs):
        if 'request' in kwargs:
            request = kwargs['request']
            del kwargs['request']
            kwargs = fill_ref_models(self, request, **kwargs)
            kwargs = self.fill_defaults(request, **kwargs)
        elif get_config("MULTI_TENANCY"):
            raise InvalidDatabaseSelected("multi tenancy enabled."
                                          " Please provide the request object to handle automatically")
        super(AdharaModel, self).__init__(*args, **kwargs)

    def __str__(self):
        if hasattr(self, "name"):
            return str(self.name)
        return str(self.pk)

    class Meta:
        abstract = True

    class Adhara:
        skip_serialization = False



from django.conf import settings
import importlib
from django.db.models import ForeignKey, OneToOneField

setting_name = "ADHARA"

try:
    SCOPE = getattr(settings, setting_name)
except AttributeError:
    SCOPE = None


def _get(lookup_scope, path_list):
    try:
        path_var = path_list.pop(0)
        lookup_scope = lookup_scope[path_var]
        if len(path_list):
            return _get(lookup_scope, path_list)
        else:
            return lookup_scope
    except KeyError:
        return None


def get_config(config_path):
    if not SCOPE:
        return None
    return _get(SCOPE, config_path.split("."))


def get_user_model():
    name = get_config("USER_MODEL.MODULE")
    package = get_config("USER_MODEL.PACKAGE")
    if not name or not package:
        name = "User"
        package = "django.contrib.auth.models"
    return getattr(importlib.import_module(package), name)


def get_user_id_field():
    return get_config("USER_MODEL.ID_FIELD") or "user_id"


def is_reference_field(field):
    # TODO check whether need to handle any other type of fields
    return type(field) in (ForeignKey, OneToOneField)



from .base import AdharaEventHandler
import requests
import json


class FirebaseEventHandler(AdharaEventHandler):

    def __init__(self, server_key):
        self.url = "https://fcm.googleapis.com/fcm/send"
        self.server_key = server_key
        super(FirebaseEventHandler, self).__init__()

    def execute(self, method, data):
        headers = {
            "Authorization": "key=" + self.server_key,
            "Content-Type": "application/json"
        }
        requests.post(self.url + (method if method is not None else ""), json.dumps(data), headers=headers)

    def notify(self, unique_id, title, content, notification_params=None, data=None):
        notification = {
            "title": title,
            "body": content
        }
        if notification_params:
            notification.update(notification_params)
        notification_object = {
            "to": unique_id,
            "notification": notification,
            "data": data
        }
        self.execute(None, notification_object)



from django.views.decorators.http import require_http_methods
from .decorators import allowed_methods, is_logged_in
from .request import APIMethods
from .restview import RestView
from .models.event_models import FirebaseEvents
from django.views.decorators.csrf import ensure_csrf_cookie
from .response_utils import ApiResponse


@require_http_methods(["GET"])
@ensure_csrf_cookie
def get_csrf_token(request):
    from django.middleware.csrf import get_token
    return ApiResponse.success(get_token(request))


@is_logged_in
@allowed_methods({APIMethods.POST, APIMethods.DELETE})
class FirebaseView(RestView):

    model = FirebaseEvents

    def pre_process(self, api_request):
        api_request.get_meta().set_required_fields("registration_token")

    def process(self):
        if self._request.get_method() == APIMethods.POST:
            return self.put_or_create()
        return super(FirebaseView, self).process()



import datetime
import json
import logging
from urllib.parse import unquote

from django.middleware.csrf import CsrfViewMiddleware
from django.utils.deprecation import MiddlewareMixin

from .exceptions import AdharaException
from .meta import Meta, Page
from .request import AdharaRequest, APIMethods
from .response_utils import ApiResponse
from .thread_local import ThreadLocal

logger = logging.getLogger(__name__)


class ThreadLocalMiddleware(MiddlewareMixin):
    """
     Simple middleware that adds the request object in thread local storage.
     Usage not recommended!
    """
    def process_request(self, request):
        ThreadLocal.set_request(request)

    def process_response(self, request, response):
        ThreadLocal._del_request()
        return response

    def process_exception(self, request, exception):
        ThreadLocal._del_request()


class DebugMiddleware(MiddlewareMixin):

    def process_request(self, request):
        request.ADHARA_DEBUG = {
            "start_time": datetime.datetime.now()
        }

    def process_response(self, request, response):
        print(
            "request {0} took {1} ms".format(
                request.path,
                datetime.datetime.now()-request.ADHARA_DEBUG["start_time"]
            )
        )
        return response


class AdharaMiddleware(MiddlewareMixin):

    def __call__(self, request):

        path = request.path
        app = path.strip('/').split('/')[0]

        try:
            self.process_request(request)
        except ValueError as ve:
            print(ve)
            return ApiResponse.error("invalid json received")
        except TypeError as te:
            print(te)
            return ApiResponse.error("input format not supported")
        except StopIteration as se:
            print(request.method, "not allowed")
            return ApiResponse.error("Method not allowed")
        except AdharaException as ae:
            return ApiResponse.error(ae.get_message())
        return self.get_response(request)

    @staticmethod
    def process_request(request):
        ct = request.META['CONTENT_TYPE']
        input_json = {}
        if request.method == 'GET':
            qs = unquote(request.META['QUERY_STRING'])
            query_params = {}
            qs = qs.split('&')
            if not ((len(qs) == 1) and (qs[0].strip() == '')):
                for q in qs:
                    q = q.split('=')
                    try:
                        query_params[q[0]] = json.loads(q[1])
                    except ValueError:
                        query_params[q[0]] = q[1]
                input_json = query_params
        if request.method == 'POST':
            if ct == "application/json":
                rb = request.body.decode('utf-8')
                if rb.strip():
                    input_json = json.loads(rb)
            elif ct == "application/x-www-form-urlencoded":
                input_json = json.loads(json.dumps(request.POST))
            elif ct.startswith("multipart/form-data"):
                pass
            else:
                raise TypeError
        if request.method == 'PUT':
            if ct == "application/json":
                rb = request.body.decode('utf-8')
                if rb.strip():
                    input_json = json.loads(rb)
            else:
                raise TypeError
        if request.method == 'DELETE':
            if ct == "application/json":
                rb = request.body.decode('utf-8')
                if rb.strip():
                    input_json = json.loads(rb)
            else:
                raise TypeError

        if input_json and 'csrfmiddlewaretoken' in input_json:
            del input_json['csrfmiddlewaretoken']
        method = getattr(APIMethods, APIMethods(request.method).name)
        meta = None
        if "_meta" in input_json:
            meta = Meta(current_page=Page(page=input_json["_meta"]["page"]))
        request.adhara_request = AdharaRequest(request, method=method, input_json=input_json, meta=meta)
        return None


class AdharaTokenMiddleware(CsrfViewMiddleware):

    def process_view(self, request, callback, callback_args, callback_kwargs):
        if request.META.get('HTTP_APIKEY'):
            return
        else:
            return super(AdharaTokenMiddleware, self).process_view(request, callback, callback_args, callback_kwargs)



from .session import Session
from .meta import Meta
from enum import Enum


class APIMethods(Enum):
    GET = "GET"
    GET_LIST = "GET_LIST"
    GET_COUNT = "GET_COUNT"
    GET_EXISTS = "GET_EXISTS"
    POST = "POST"
    PUT = "PUT"
    BULK_PUT = "BULK_PUT"
    DELETE = "DELETE"
    BULK_DELETE = "DELETE"


class Criteria:

    def __init__(self, criteria=None, identifier=None):
        self._criteria = criteria or {}
        self._identifier = identifier

    def update_identifier(self, identifier):
        self._identifier = identifier

    def set_criteria(self, criteria=None):
        self._criteria = criteria or {}

    def get_criteria(self):
        if self._identifier:
            criteria = {"pk": self._identifier}
            criteria.update(self._criteria)
        else:
            criteria = self._criteria
        return criteria

    def update_criteria(self, criteria):
        self._criteria.update(criteria)


class AdharaRequest:

    def __init__(self, request, model=None, method=None, input_json=None, identifier=None,
                 meta=None, criteria=None, exclude_criteria=None, response_key=None, transformer=None):
        self._request = request
        self._model = model
        self._identifier = identifier
        self._input_json = input_json
        self._identifier = identifier
        self._method = None
        self._set_method_(method)
        self._meta = meta or Meta()
        self._criteria = Criteria(criteria, identifier)
        self._exclude_criteria = Criteria(exclude_criteria)
        self._sub_requests = []
        self._response_key = response_key
        self.transformer = transformer
        self.is_sub_request = False
        self.model_manager = None

    def _get_method_for(self, method, identifier):
        if method in [APIMethods.GET, APIMethods.GET_LIST]:
            return APIMethods.GET if identifier else APIMethods.GET_LIST
        if method in [APIMethods.DELETE, APIMethods.BULK_DELETE]:
            return APIMethods.DELETE if identifier else APIMethods.BULK_DELETE
        if method in [APIMethods.PUT, APIMethods.BULK_PUT]:
            return APIMethods.PUT if identifier else APIMethods.BULK_PUT
        return method

    def _set_method_(self, method=None):
        self.set_method(self._get_method_for(method or self.get_method(), self.get_identifier()))

    def get_method(self):
        return self._method

    def set_method(self, method):
        self._method = method

    def get_criteria_object(self):
        return self._criteria

    def set_exclude_criteria(self, exclude_criteria):
        if isinstance(exclude_criteria, Criteria):
            self._exclude_criteria = exclude_criteria
        else:
            self._exclude_criteria = Criteria(exclude_criteria, self.get_identifier())

    def get_exclude_criteria_object(self):
        return self._exclude_criteria

    def get_meta(self):
        return self._meta

    def get_input(self):
        required_input = {
            "request": self._request
        }
        if self._input_json:
            required_input.update(self._input_json)
        if 'id' in required_input:
            del required_input['id']
        return required_input

    def update_input(self, update_data):
        if not self._input_json:
            self._input_json = {}
        self._input_json.update(update_data)

    def get_identifier(self):
        return self._identifier

    def set_identifier(self, identifier=None):
        self._identifier = identifier
        if self.get_criteria_object():
            self.get_criteria_object().update_identifier(identifier)
        if self.get_exclude_criteria_object():
            self.get_exclude_criteria_object().update_identifier(identifier)
        self._set_method_()

    def set_new_identifier(self, new_identifier):
        self._identifier = new_identifier

    def get_tenant_db(self):
        return Session.get_db(self._request)

    def get_http_request(self):
        return self._request

    def add_sub_request(self, sub_request):
        sub_request.is_sub_request = True
        self._sub_requests.append(sub_request)

    def get_sub_requests(self):
        return self._sub_requests

    def set_model(self, model):
        self._model = model

    def get_model(self):
        return self._model

    def get_model_manager(self):
        if self.is_sub_request:
            return self.model_manager
        return self.get_model().objects

    def get_response_key(self):
        return self._response_key



from collections import namedtuple

from django.conf import settings

from .request import APIMethods
from .response_utils import ApiResponse
from .session import Session


def allowed_methods(api_methods):
    """
    Decorator to make a view only accept particular request methods.  Usage::

        @require_http_methods({adhara.request.APIMethods.GET, adhara.request.APIMethods.POST})
        class XViewClass(RestView):
            # I can assume now that only GET (unique) or POST requests make it this far
            # ...

    Note that request methods should be passing a set of `adhara.request.APIMethods` types
    :param {Set<adhara.request.APIMethods>} api_methods: set of allowed Adhara API methods
    """
    all_methods = set(APIMethods)
    restricted_methods = all_methods.difference(api_methods)

    def _method_not_allowed(self):
        self._response.error({"allowed_methods": list(map(lambda x: x.value, api_methods))}, error_code=403)
        return self._response
        # pratikriyaa = {
        #     "status": "error",
        #     "data":
        # }
        # return ApiResponse.NotAllowedPratikriyaa(pratikriyaa)
        # return HttpResponseNotAllowed(list(map(lambda x:x.value, api_methods)))

    def fn(cls):
        for method in restricted_methods:
            setattr(cls, method.value.lower(), _method_not_allowed)
        return cls
    return fn


def is_logged_in(cls):
    default_dispatcher = cls.dispatch

    def auth_and_dispatch(self, request, *args, **kwargs):
        if Session.is_logged_in(request):
            return default_dispatcher(self, request, *args, **kwargs)
        else:
            return ApiResponse.error("Not Logged In")
    cls.dispatch = auth_and_dispatch
    return cls


StaticFiles = namedtuple("StaticFiles", ["CSS", "JS"])
_CSS = settings.ADHARA["CSS"] if "CSS" in settings.ADHARA else {}
_JS = settings.ADHARA["JS"] if "JS" in settings.ADHARA else {}


def _get_for_context(contexts):
    if type(contexts) is str:
        contexts = [contexts]
    css_files = []
    js_files = []
    for context in contexts:
        css_files.extend(_CSS[context])
        js_files.extend(_JS[context])
    return StaticFiles(css_files, js_files)


def include_static_contexts(contexts):
    def ret(fn):
        def req_incl_fn(request, *args, **kwargs):
            kwargs['static_files'] = _get_for_context(contexts)
            return fn(request, *args, **kwargs)
        return req_incl_fn
    return ret



from django.http import JsonResponse


class ApiResponse:

    class Pratikriyaa(JsonResponse):
        """
        Error Response Class
        """
        pass

    class SkhalitaPratikriyaa(JsonResponse):
        """
        Error Response Class
        """
        status_code = 400

        def __init__(self, data, **kwargs):
            super(ApiResponse.SkhalitaPratikriyaa, self).__init__(data, safe=False, **kwargs)

    class AsiddhauPratikriyaa(JsonResponse):
        """
        Failure Response Class
        """
        status_code = 500

        def __init__(self, data, **kwargs):
            super(ApiResponse.AsiddhauPratikriyaa, self).__init__(data, safe=False, **kwargs)

    class NotAllowedPratikriyaa(JsonResponse):
        """
                Failure Response Class
                """
        status_code = 405

        def __init__(self, data, **kwargs):
            super(ApiResponse.NotAllowedPratikriyaa, self).__init__(data, safe=False, **kwargs)

    @staticmethod
    def success(data=None, meta=None):
        pratikriyaa = {
            "status": "success",
            "data": data
        }
        if meta is not None:
            pratikriyaa["meta"] = meta
        return ApiResponse.Pratikriyaa(pratikriyaa)

    @staticmethod
    def error(message):
        pratikriyaa = {
            "status": "error",
            "data": message
        }
        return ApiResponse.SkhalitaPratikriyaa(pratikriyaa)

    @staticmethod
    def failure(failures):
        pratikriyaa = {
            "status": "failure",
            "data": failures
        }
        return ApiResponse.AsiddhauPratikriyaa(pratikriyaa)



from threading import local
_thread_locals = local()


class ThreadLocal:

    @staticmethod
    def set_request(request):
        _thread_locals.request = request

    @staticmethod
    def _del_request():
        if hasattr(_thread_locals, 'request'):
            del _thread_locals.request

    @staticmethod
    def get_request():
        return _thread_locals.getattr("request")



class AdharaException(Exception):

    def get_message(self):
        return self.args[0]


class InvalidInput(AdharaException):
    pass


class InvalidOperationException(AdharaException):
    pass


class InvalidDatabaseSelected(AdharaException):
    pass


class UnAuthorizedException(AdharaException):
    pass



from django.conf.urls import url
from . import views


urlpatterns = [
    url(r'^token$', views.get_csrf_token, name="token"),
    url(r'^events/firebase$', views.FirebaseView.as_view(), name="firebase_registration")
]



from itertools import chain
import importlib
from .exceptions import AdharaException
from .request import AdharaRequest, APIMethods
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.db import utils as django_utils
from .utilities import is_reference_field
from django.views.generic import View

from .models.base_models import fill_ref_models
from .response import AdharaResponse, ResponseStatus
from .session import Session
from .utilities import get_user_model
import logging

logger = logging.getLogger("adhara.restview.RestView")


class RestView(View):
    """
    Rest view to handle API responses in a REST Pattern.
    _meta pattern for get_list
    _meta : {
        "page": {
            "start" : "1",
            "end": "10"
        }
    }
    """

    model = None

    def __init__(self, adhara_request=None, **kwargs):
        self._request = None
        self._db = None
        self._response = None
        if isinstance(adhara_request, AdharaRequest):
            self.set_api_request(adhara_request)
        super(RestView, self).__init__(**kwargs)

    def set_api_request(self, adhara_request):
        self._request = adhara_request
        if not getattr(adhara_request, "transformer", None):
            adhara_request.transformer = self._transform_row
        self._db = adhara_request.get_tenant_db()
        self._response = AdharaResponse(self._request)

    def _get_query_criteria(self):
        return self._request.get_criteria_object().get_criteria()

    def _get_exclude_criteria(self):
        return self._request.get_exclude_criteria_object().get_criteria()

    def _get_object(self, *fields):
        criteria = self._request.get_criteria_object().get_criteria()
        try:
            if len(fields):
                return self._request.get_model_manager().values(*fields).using(self._db).get(**criteria)
            else:
                return self._request.get_model_manager().using(self._db).get(**criteria)
        except ObjectDoesNotExist:
            return None

    def pre_authorize(self, request):
        """
        any validations like licensing check etc to be handled here
        :param {AdharaRequest} request: adhara request
        :return:
        """
        pass

    def _pre_process(self, *args, **kwargs):
        args_from_path = {}
        for key, value in kwargs.items():
            if key == "pk":
                self._request.set_identifier(kwargs["pk"])
            elif key.startswith("q_"):
                args_from_path[key[2:]] = value

        if not self._request.get_model():
            self._request.set_model(self.model)

        user_model = get_user_model()
        opts = self._request.get_model()._meta
        for f in chain(opts.concrete_fields, opts.private_fields, opts.many_to_many):
            if is_reference_field(f) and user_model == f.model._meta.get_field(f.name).rel.to:
                try:
                    self._request.get_input()[f.name]
                except KeyError:
                    if self._request.get_method() in [APIMethods.POST, APIMethods.PUT]:
                        args_from_path[f.name] = Session.get_user_id(self._request.get_http_request())

        if args_from_path:
            if self._request.get_method() in [APIMethods.POST]:
                self._request.update_input(args_from_path)
            self._request.get_criteria_object().update_criteria(criteria=args_from_path)

    def pre_process(self, api_request):
        """
        modify the request bu adding sub requests, or change the input data, etc.
        :param api_request: adhara.request.AdharaRequest
        :return: None
        """
        pass

    def authorize_crud(self, method, instance):
        """
        Authorize CRUD methods on the current view
        :param method: APIMethod
        :param instance: model entry
        :return:{bool} whether method is authorized or not
        """
        pass

    def authorize(self, request, instance=None, parent_instance=None):
        """
        :raises:{AdharaException}
        """
        pass

    def process(self):
        """
        executes related http method and returns an AdharaResponse object
        :return:{AdharaRequest}
        """
        return getattr(self, self._request.get_method().value.lower())()

    def _transform_row(self, row, transformed_row):
        self._get_permissions(row)
        self.transform_row(row, transformed_row)

    def transform_row(self, row, transformed_row):
        """
        modify a row data which will be returned in the response
        :param row: Django db entry instance
        :param transformed_row: Transformed row data in dict format
        :return: None
        """
        pass

    def _post_process(self):
        """
        holder to add code for post processing by rest view.
        Currently handle's adding permissions
        :return: None
        """
        # self._get_permissions()
        pass

    def post_process(self, response, success):
        """
        can handle methods after processing current adhara request
        :param response: AdharaResponse
        :return:None
        """
        pass

    def _get_permissions(self, instance):
        """
        Constructs permissions and sets it to response object
        :return:None
        """
        # Post.like_set.rel.related_model
        # Post._meta.related_objects[0].related_model.Adhara.rest_view
        for operation in (APIMethods.PUT, APIMethods.DELETE):
            self._response.set_permission(
                instance.pk,
                "edit" if operation == APIMethods.PUT else operation.value.lower(),
                self.authorize_crud(operation, instance))

        for ro in self._request.get_model()._meta.related_objects:
            try:
                related_rest_view = ro.related_model.Adhara.rest_view
                if not related_rest_view:
                    continue
            except AttributeError:
                continue
            splits = related_rest_view.split(".")
            rest_view_class = getattr(importlib.import_module(".".join(splits[0:-1])), splits[-1])
            related_field_name = ro.related_model._meta.object_name.lower()
            related_adhara_request = \
                AdharaRequest(self._request.get_http_request(), model=ro.related_model, method=APIMethods.POST)
            related_rv = rest_view_class(related_adhara_request)
            try:
                related_rv.authorize(related_adhara_request, parent_instance=instance)
                self._response.set_permission(instance.pk, related_field_name, True)
            except AdharaException:
                self._response.set_permission(instance.pk, related_field_name, False)

    def _extract_required_fields(self, instance):
        opts = instance._meta
        required_fields = self._request.get_meta().get_required_fields()
        for f in chain(opts.concrete_fields, opts.private_fields, opts.many_to_many):
            if required_fields and f.name not in required_fields:
                try:
                    delattr(instance, f.name)
                except AttributeError as ae:
                    logger.debug(ae)
        return instance

    def get_count(self):
        self._response.success(
            self._request.get_model_manager().using(self._db).filter(**self._get_query_criteria()).count())
        return self._response

    def get_exists(self):
        self._response.success(self._get_object("pk") is not None)
        return self._response

    def post(self):
        with transaction.atomic(using=self._db):
            input_data = self._request.get_input()
            entry = self._request.get_model()(**input_data)
            self.authorize(self._request, instance=entry)
            try:
                entry.save(using=self._db)
            except django_utils.DatabaseError as e:
                self._response.error(e.args[1])
                return self._response
            # self._request.NEW_PRIMARY_KEY = entry.pk    # TODO
        self._response.success(self._extract_required_fields(entry))
        return self._response

    def get(self):
        entry = self._get_object(*self._request.get_meta().get_required_fields())
        self.authorize(self._request, instance=entry)
        if entry is None:
            self._response.error("Invalid identifier")
        else:
            self._response.success(entry)
        return self._response

    def get_list(self):
        # TODO required fields
        qs = self._request.get_model_manager().using(self._db)\
            .exclude(**self._get_exclude_criteria())\
            .filter(**self._get_query_criteria())
        order_by = self._request.get_meta().get_order_by()
        if order_by:
            qs = qs.order_by(*order_by)
        total_count = len(qs)
        meta = self._request.get_meta()
        pagination = meta.get_pagination()
        curr_page = pagination.get_current_page()
        qs = qs[curr_page.get_start()-1: curr_page.get_end()]
        count = len(qs)
        meta.update_page_properties(count, total_count)
        for q in qs:
            self.authorize(self._request, instance=q)
        self._response.success(qs)
        return self._response

    def _put(self, entry):
        self.authorize(self._request, instance=entry)
        filled_input = fill_ref_models(entry, **self._request.get_input())
        for key, value in filled_input.items():
            setattr(entry, key, value)
        try:
            entry.save()
        except django_utils.DatabaseError as e:
            self._response.error(e.args[1])
        self._response.success(self._extract_required_fields(entry))
        return self._response

    def put(self):
        with transaction.atomic(using=self._db):
            entry = self._get_object()
            if entry is None:
                self._response.error("Does not exist")
            return self._put(entry)

    def put_or_create(self):
        with transaction.atomic(using=self._db):
            entry = self._get_object()
            if entry is None:
                return self.post()
            return self._put(entry)

    def delete(self):
        try:
            entry = self._request.get_model_manager().using(self._db).get(**self._get_query_criteria())
            self.authorize(self._request, instance=entry)
            entry.delete()
            self._response.success("Deleted Successfully")
        except django_utils.DatabaseError as e:
            self._response.error(e.args[1])
        return self._response

    def bulk_delete(self):
        try:
            entries = self._request.get_model_manager().using(self._db).filter(**self._get_query_criteria())
            for entry in entries:
                self.authorize(self._request, instance=entry)
            entries.delete()
            self._response.success("Deleted Successfully")
        except django_utils.DatabaseError as e:
            self._response.error(e.args[1])
        return self._response

    def execute(self, *args, **kwargs):
        try:
            self.pre_authorize(self._request)
        except AdharaException:
            self._response.error("Unauthorized")
            return self._response.serialize()
        try:
            self._pre_process(*args, **kwargs)
            self.pre_process(self._request)
            response = self.process()
            response.transform_data()
            self._post_process()
            self.post_process(self._response, self._response.get_response_status() == ResponseStatus.SUCCESS)
        except AdharaException as ae:
            self._response.error(ae.get_message())
            return self._response.serialize()
        return response.serialize()

    def dispatch(self, request, *args, **kwargs):
        self.set_api_request(request.adhara_request)
        return self.execute(*args, **kwargs)



from ..models.event_models import FirebaseEvents
from ..utilities import get_config


def emit(user_id, title, content, notification_params, data):
    events = get_config("EVENTS")
    for event_type, config in events.items():
        if event_type == "FIREBASE":
            from .firebase import FirebaseEventHandler
            firebase = FirebaseEventHandler(config["SERVER_KEY"])
            user_firebase_config = FirebaseEvents.objects.values("registration_token").get(user_id=user_id)
            firebase.notify(user_firebase_config["registration_token"], title, content, notification_params, data)
        else:
            pass



from django.contrib.auth import logout as dj_logout
from .response_utils import ApiResponse
from .utilities import get_user_id_field
from django.conf import settings

AUTH_TOKEN_HEADER = "HTTP_APIKEY"


class Session:
    ADHARA_LOGGED_IN = "adhara_logged_in"
    USING_DB = "using_db"
    USER_ID = "user_id"

    @staticmethod
    def set_variable(request, key, value):
        if 'adhara' not in request.session:
            request.session['adhara'] = {}
        request.session['adhara'][key] = value

    @staticmethod
    def get_variable(request, key):
        # If any changes made here, must also be changed in middleware
        # as the same verification method is re written in middleware
        try:
            return request.session['adhara'][key]
        except KeyError:
            return Session.get_token(request, key)

    @staticmethod
    def set_token(request, key, value):
        if hasattr(request, 'adhara_authorization'):
            request.adhara_authorization = {}
        request.adhara_authorization[key] = value

    @staticmethod
    def get_token(request, key):
        try:
            return request.adhara_authorization[key]
        except AttributeError:
            return None

    @staticmethod
    def get_db(request):
        return Session.get_variable(request, Session.USING_DB)

    @staticmethod
    def get_user_id(request):
        return Session.get_variable(request, Session.USER_ID)

    @staticmethod
    def login(request, user_id=None, db=None):
        Session.set_variable(request, Session.ADHARA_LOGGED_IN, True)
        Session.set_variable(request, Session.USER_ID, user_id)
        Session.set_variable(request, Session.USING_DB, db)

    @staticmethod
    def token_login(request, user_id=None, db=None):
        Session.set_token(request, Session.ADHARA_LOGGED_IN, True)
        Session.set_token(request, Session.USER_ID, user_id)
        Session.set_token(request, Session.USING_DB, db)

    @staticmethod
    def is_logged_in(request):
        return Session.get_variable(request, Session.ADHARA_LOGGED_IN)

    @staticmethod
    def logout(request):
        dj_logout(request)


def is_logged_in(fn):

    def logged_in(request, *args, **kwargs):
        if Session.is_logged_in(request):
            return fn(request, *args, **kwargs)
        else:
            return ApiResponse.error("Not Logged In")
    return logged_in


def get_user_criteria(request):
    return {get_user_id_field(): Session.get_user_id(request)}


def fill_user_id(fn):

    def fnr(request, *args, **kwargs):
        if request.method == "POST":
            request.INPUT_JSON.update(get_user_criteria(request))
        return fn(request, *args, **kwargs)
    return fnr



from .base_models import AdharaModel
from django.db import models
from ..utilities import get_user_model, get_config


events_list = get_config("EVENTS.INCLUDE")


def is_abstract(event_type):
    if events_list:
        try:
            events_list.index(event_type)
            return False
        except ValueError:
            return True
    else:
        return True


class AdharaEvents(AdharaModel):
    user = models.OneToOneField(get_user_model(), null=False)

    class Meta:
        abstract = True
        app_label = "adhara"


class FirebaseEvents(AdharaEvents):
    registration_token = models.CharField(max_length=255)

    class Meta:
        db_table = "firebase_events"
        abstract = is_abstract("FIREBASE")



from django.test import TestCase
import inspect
from django.db.models import Model as DjangoBaseModel
from adhara.models import AdharaModel


class ExtendsAdharaModel(TestCase):

    def __init__(self, models, method_name):
        self.models = models
        super(ExtendsAdharaModel, self).__init__(method_name)

    def test_models_in_app(self):
        if not self.models:
            return
        for name, obj in inspect.getmembers(self.models):
            if inspect.isclass(obj) and issubclass(obj, DjangoBaseModel):
                extends_adhara_model = issubclass(obj, AdharaModel)
                if not extends_adhara_model:
                    print(str(obj), "doesn't extend Adhara model")
                assert extends_adhara_model
